/*
 * Copyright 2012, 2016, 2017 Sony Corporation
 */
#include <linux/module.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <linux/irq.h>
#include <linux/interrupt.h>
#include <linux/kthread.h>
#include <linux/slab.h>
#include <linux/delay.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/mmc/card.h>
#include <linux/mmc/host.h>
#include <mach/icx_carddet.h>
#include <mach/mt_gpio.h>
#include <mach/board.h>
#include <mach/eint.h>
#include <cust_eint.h>

#include "mt_sd.h"
#include "drivers/mmc/card/queue.h"

/****************/
/*@ Definitions */
/****************/
#define ENABLE_STAT_QUEUE
#define MODULE_NAME         "icx_carddet"

#ifdef BENTEN_M1_SD_POWER_SWITCH_GPIO
/* Not supported (Only use for debug) */
/* #define POWER_CONTROL_ENABLE_BY_INTERNAL */
#define POWER_CONTROL_ENABLE_BY_EXTERNAL
#endif /* BENTEN_M1_SD_POWER_SWITCH_GPIO */

/* for debug */
#define CARDDET_DEBUG   0

#define ERR_MSG(format, arg...)  pr_err("%s: ** Error ** " format "\n", \
					 MODULE_NAME, ## arg)
#define INFO_MSG(format, arg...) pr_info("%s: " format "\n", \
					 MODULE_NAME, ## arg)

#if CARDDET_DEBUG
#define DBG_MSG(format, arg...)  pr_warn("%s: " format "\n", \
					 MODULE_NAME, ## arg)
#else  /* CARDDET_DEBUG */
#define DBG_MSG(format, arg...)  do {} while (0)
#endif /* CARDDET_DEBUG */


/* Major & Minor */
#define CARDDET_DRV_MAJOR            0        /* 0:dynamic */
#define CARDDET_DRV_MINOR            0
#define CARDDET_CDEV_FILE_NAME       "icx_carddet"

/* PIN assign */
#define CARD_SD_DET                 1
#define CARD_MS_DET                 1
#define CARD_PWRCTL                 1
#define CHG_PINSEL_SD_SD            0x00000000
#define CHG_PINSEL_SD_MS            0x00000001

/* Request ID to thread */
enum {
	CARDDET_REQ_NONE = 0,
	CARDDET_REQ_IRQ,
	CARDDET_REQ_INIT,
};

/* Sequence of thread */
enum {
	CARDDET_SEQ_NONE = 0,
	CARDDET_SEQ_PWRON,
	CARDDET_SEQ_SET_PULL,
	CARDDET_SEQ_INIT,
	CARDDET_SEQ_RETRY,
};

/* Timeout */
#define CARDDET_TIMEOUT_IDLE                   0    /* none   */
#define CARDDET_TIMEOUT_WAIT_PWR_STABLE        100  /*   1  s */
#define CARDDET_TIMEOUT_WAIT_PULL_STABLE       1    /*  10 ms */
#define CARDDET_TIMEOUT_WAIT_INIT              90   /* 900 ms */
#define CARDDET_TIMEOUT_RETRY                  75   /* 750 ms */

/* Retry Count */
#define    CARDDET_RETRY_OVER           3

/* Sub Status */
#define CARDDET_STATE_FLAG_NONE         0x00
#define CARDDET_STATE_FLAG_WAIT_STABLE  0x01
#define CARDDET_STATE_FLAG_WAIT_REQUEST 0x02
#define CARDDET_STATE_FLAG_ALL          0x03

/***************/
/*@ Structures */
/***************/
/* CARDDET device */
struct carddet_dev {
	int                     users;
	struct cdev             cdev;
	spinlock_t              lock;
	wait_queue_head_t       wqh;
	int                     irq;
	struct task_struct      *thread_task;
	wait_queue_head_t       thread_wq;
	int                     thread_req;
	int                     thread_seq;
	int                     thread_timeout;
	int                     det_state;
	int                     card_type;
	int                     read_only;
	int                     retry_count;
	int                     suspend_flag;
	unsigned char           insert_state_flag;
#ifdef ENABLE_STAT_QUEUE
	struct list_head     list;
#else  /* ENABLE_STAT_QUEUE */
	int                     wake_up_flg;
#endif /* ENABLE_STAT_QUEUE */
	struct msdc_host        *host_sd;
};

#ifdef ENABLE_STAT_QUEUE
struct carddet_detect_status {
	struct list_head entry;
	int              state;
	int              type;
	int              ro;
};
#endif /* ENABLE_STAT_QUEUE */

struct mmc_blk_data {
	spinlock_t lock;
	struct gendisk *disk;
	struct mmc_queue queue;

	unsigned int usage;
	unsigned int read_only;
};

static struct class *carddet_class;

/********************/
/*@ Local variables */
/********************/
static DEFINE_SEMAPHORE(carddet_sem);

/* Major & Minor */
static int carddet_major = CARDDET_DRV_MAJOR;
static int carddet_minor = CARDDET_DRV_MINOR;


module_param(carddet_major, int, S_IRUGO);
module_param(carddet_minor, int, S_IRUGO);


static int prev_det_state;
static int enable_req_det_func;

#if CARDDET_DEBUG
module_param(enable_req_det_func, int, S_IRUGO | S_IWUSR);
#endif


/* CARDDET devices */
static struct carddet_dev *carddet_devices;

/***************/
/*@ Prototypes */
/***************/
/* driver entry routines */
static irqreturn_t carddet_eint_handler(int irq, void *arg);
static void carddet_detect_card_removed(void);
static int  carddet_get_card_state(void);

/* user entry routines */
static int carddet_open(struct inode *inode, struct file *filp);
static int carddet_release(struct inode *inode, struct file *filp);
static int carddet_ioctl(struct file *filp, unsigned int cmd,
			unsigned long arg);
static unsigned int carddet_poll(struct file *file, poll_table *wait);

/* file operations */
static const struct file_operations carddet_fops = {
	.owner   = THIS_MODULE,
	.open    = carddet_open,
	.release = carddet_release,
	.unlocked_ioctl   = carddet_ioctl,
	.poll    = carddet_poll,
};

/* Local routines */
static int  carddet_notify_status(struct carddet_dev *dev);
static void carddet_control_power(int sts);
static int  carddet_check_result(void);
static int  carddet_ioc_get_info(struct carddet_dev *dev, unsigned long arg);
static int  carddet_ioc_control_power(struct carddet_dev *dev,
				      unsigned long arg);
static int  carddet_ioc_get_xdet(struct carddet_dev *dev, unsigned long arg);
static int  carddet_ioc_req_detect(struct carddet_dev *dev);

static struct carddet_eint_ops carddet_ops = {
	.notifier_call = carddet_eint_handler,
};

static struct carddet_func_tbl carddet_func = {
	.card_removed = carddet_detect_card_removed,
};

static struct carddet_to_msc_func carddet_to_msc_func_tbl = {
	.get_status = carddet_get_card_state,
};


/*******************/
/*@ LOCAL ROUTINES */
/*******************/
/* Get card status(for msc module) */
static int carddet_get_card_state(void)
{
	int ret = 0;

	DBG_MSG("** %s called", __func__);

	if (carddet_devices)
		ret = carddet_devices->det_state;

	return ret;
}

/* Notify the card status to user */
static int carddet_notify_status(struct carddet_dev *dev)
{
	int ret = 0;
#ifdef ENABLE_STAT_QUEUE
	struct carddet_detect_status *status;
#endif

	DBG_MSG("** %s called", __func__);

#ifdef ENABLE_STAT_QUEUE
	status = kmalloc(sizeof(struct carddet_detect_status), GFP_KERNEL);
	if (status) {
		status->state = dev->det_state;
		status->type  = dev->card_type;
		status->ro    = dev->read_only;

		list_add_tail(&status->entry, &dev->list);

		wake_up_interruptible(&dev->wqh);
	} else {
		ret = -ENOMEM;
	}
#else  /* ENABLE_STAT_QUEUE */
	if (dev->wake_up_flg == 0) {
		dev->wake_up_flg = 1;

		wake_up_interruptible(&dev->wqh);
	}
#endif /* ENABLE_STAT_QUEUE */

	return ret;
}

/* Power control */
static void carddet_control_power(int sts)
{
#ifdef POWER_CONTROL_ENABLE_BY_INTERNAL
	int ret = 0;
	int gpio;

	if (sts == ICX_CARDDET_STATE_INSERTED)
		gpio = GPIO_OUT_ONE;
	else if (sts == ICX_CARDDET_STATE_REMOVED)
		gpio = GPIO_OUT_ZERO;
	else
		gpio = GPIO_OUT_ZERO;

	ret = mt_set_gpio_out(GPIO111, gpio);
	DBG_MSG("** %s called ret(%d) gpio[%d](%d) sts(%d)",
		 __func__, ret, GPIO111, gpio, sts);

#else  /* POWER_CONTROL_ENABLE_BY_INTERNAL */
	/* No operation for now */
	DBG_MSG("** %s called sts(%d) !!! NOP !!!", __func__, sts);
#endif /* POWER_CONTROL_ENABLE_BY_INTERNAL */
}

static int carddet_get_hw_detect_state(void)
{
	int detect_status = 0;
	unsigned int polarity = 0;
	struct msdc_host *host = NULL;

	host = carddet_devices->host_sd;
	if (host ==  NULL) {
		DBG_MSG("** %s called", __func__);
		return 0;
	}
#ifdef CUST_EINT_MSDC1_INS_NUM
	polarity = mt_eint_get_polarity(CUST_EINT_MSDC1_INS_NUM);
#endif
	if (polarity ^ host->hw->cd_level)
		detect_status = 1;

	DBG_MSG("** %s called detect_status(%d)", __func__, detect_status);

	return detect_status;
}

/* get driver data */
static int carddet_check_result(void)
{
	int           result;
	unsigned long flags;
	int ro  = 0;
	int ret = 0;
	struct msdc_host *host_sd;
	struct mmc_blk_data *md = NULL;

	DBG_MSG("** %s called", __func__);
	host_sd = carddet_devices->host_sd;

	if (host_sd->mmc == NULL || host_sd->mmc->card == NULL) {
		DBG_MSG("** %s mmc or card == NULL ", __func__);
		result = ICX_CARDDET_RESULT_NG;
	} else {
		md = mmc_get_drvdata(host_sd->mmc->card);
		if (md != NULL) {
			result = ICX_CARDDET_RESULT_OK;
			ro = (host_sd->mmc->card->state & MMC_STATE_READONLY);
		} else {
			result = ICX_CARDDET_RESULT_NG;
		}
	}

	if (result == ICX_CARDDET_RESULT_OK) {
		DBG_MSG("Detection completed : type=%d ro=%d",
			 carddet_devices->card_type, ro);

		spin_lock_irqsave(&carddet_devices->lock, flags);
		carddet_devices->det_state = ICX_CARDDET_STATE_DETECTED;
		carddet_devices->read_only = ro;
		spin_unlock_irqrestore(&carddet_devices->lock, flags);

		carddet_notify_status(carddet_devices);
	} else {
		if (carddet_devices->retry_count >= CARDDET_RETRY_OVER) {
			DBG_MSG("Detection failed : type=%d",
				 carddet_devices->card_type);

			spin_lock_irqsave(&carddet_devices->lock, flags);
			carddet_devices->det_state = ICX_CARDDET_STATE_NG;
			carddet_devices->read_only = 0;
			spin_unlock_irqrestore(&carddet_devices->lock, flags);

			carddet_control_power(carddet_devices->det_state);

			carddet_notify_status(carddet_devices);
		} else {
			DBG_MSG("Detection retry(%d) : type=%d",
				 carddet_devices->retry_count,
				 carddet_devices->card_type);

			carddet_devices->retry_count++;
			ret = -1;
		}
	}

	return ret;
}

/* Removed function */
static void carddet_card_removed(void)
{
	unsigned long flags;
	int card_type;

	DBG_MSG("** %s called", __func__);

	carddet_devices->thread_seq     = CARDDET_SEQ_NONE;
	carddet_devices->thread_timeout = CARDDET_TIMEOUT_IDLE;

	spin_lock_irqsave(&carddet_devices->lock, flags);
	card_type = carddet_devices->card_type;
	carddet_devices->det_state = ICX_CARDDET_STATE_REMOVED;
	carddet_devices->card_type = ICX_CARDDET_TYPE_NONE;
	carddet_devices->read_only = 0;
	carddet_devices->insert_state_flag = CARDDET_STATE_FLAG_NONE;

	spin_unlock_irqrestore(&carddet_devices->lock, flags);

	carddet_control_power(carddet_devices->det_state);

	if (carddet_devices->suspend_flag == 0)
		carddet_notify_status(carddet_devices);

	carddet_devices->thread_req = CARDDET_REQ_NONE;
}

static void carddet_detect_card_removed(void)
{
	DBG_MSG("** %s called from msdc driver", __func__);

	carddet_card_removed();
}

/* ioctl(CARD_DET_GET_INFO) */
static int carddet_ioc_get_info(struct carddet_dev *dev, unsigned long arg)
{
	icx_card_info_t info;
	unsigned long   flags;

	DBG_MSG("** %s called", __func__);

	spin_lock_irqsave(&dev->lock, flags);

#ifdef ENABLE_STAT_QUEUE
	if (!list_empty(&dev->list)) {
		/* pop status */
		struct carddet_detect_status *status =
			list_first_entry(&dev->list,
					 struct carddet_detect_status, entry);
		if (status) {
			info.state = status->state;
			info.type  = status->type;
			info.ro    = status->ro;

			list_del(&status->entry);
			kfree(status);
		} else {
			info.state = dev->det_state;
			info.type  = dev->card_type;
			info.ro    = dev->read_only;
		}
	} else {
		info.state = dev->det_state;
		info.type  = dev->card_type;
		info.ro    = dev->read_only;
	}
#else  /* ENABLE_STAT_QUEUE */
	info.state = dev->det_state;
	info.type  = dev->card_type;
	info.ro    = dev->read_only;
#endif /* ENABLE_STAT_QUEUE */

	spin_unlock_irqrestore(&dev->lock, flags);

	if (copy_to_user((icx_card_info_t *)arg, &info, sizeof(info)))
		return -EFAULT;

	return 0;
}

/* ioctl(CARD_DET_CONTROL_POWER) */
static int carddet_ioc_control_power(struct carddet_dev *dev, unsigned long arg)
{
	int ret  = 0;
	int gpio = 0;
	unsigned int power_mode;

	ret = copy_from_user(&power_mode,
			    (unsigned int *)arg, sizeof(power_mode));
#ifdef POWER_CONTROL_ENABLE_BY_EXTERNAL
	if (power_mode == 0) {
		/* power off */
		gpio = GPIO_OUT_ZERO;
	} else {
		/* power on */
		gpio = GPIO_OUT_ONE;
	}
	ret = mt_set_gpio_out(GPIO111, gpio);
	DBG_MSG("** %s called ret(%d) gpio[%d](%d) PM(%d)",
		__func__, ret, GPIO111, gpio, power_mode);

#else  /* POWER_CONTROL_ENABLE_BY_EXTERNAL */
	/* No operation for now */
	DBG_MSG("** %s called power_mode(%d) !!! NOP !!!",
		__func__, power_mode);
#endif /* POWER_CONTROL_ENABLE_BY_EXTERNAL */

	return 0;
}

/* ioctl(CARD_DET_GET_XDET) */
static int carddet_ioc_get_xdet(struct carddet_dev *dev, unsigned long arg)
{
	int detect_status = 0;

	DBG_MSG("** %s called", __func__);

	detect_status = carddet_get_hw_detect_state();

	if (copy_to_user((int *)arg, &detect_status, sizeof(int)))
		return -EFAULT;

	return 0;
}

/* ioctl(CARD_DET_REQ_DETECT) */
static int carddet_ioc_req_detect(struct carddet_dev *dev)
{
	unsigned long flags;
	int ret = 0;

	if (enable_req_det_func) {
		DBG_MSG("** %s called", __func__);

		spin_lock_irqsave(&carddet_devices->lock, flags);

		if (dev->insert_state_flag & CARDDET_STATE_FLAG_WAIT_REQUEST) {
			dev->insert_state_flag &=
				~CARDDET_STATE_FLAG_WAIT_REQUEST;

			if (!(dev->insert_state_flag &
					CARDDET_STATE_FLAG_WAIT_STABLE)) {
				carddet_devices->thread_req = CARDDET_REQ_INIT;
				wake_up_interruptible(
						&carddet_devices->thread_wq);
			}
		} else {
			ret = -EAGAIN;
		}

		spin_unlock_irqrestore(&carddet_devices->lock, flags);
	} else {
		DBG_MSG("** %s called", __func__);
	}

	return ret;
}

/************************/
/*@ USER ENTRY ROUTINES */
/************************/
/* OPEN */
static int carddet_open(struct inode *inode, struct file *filp)
{
	struct carddet_dev *dev;

	DBG_MSG("** %s called  :  open", __func__);

	dev = container_of(inode->i_cdev, struct carddet_dev, cdev);
	filp->private_data = dev;

	if (down_interruptible(&carddet_sem))
		return -ERESTARTSYS;

	dev->users++;
	DBG_MSG("** carddet: users=%d", dev->users);

	up(&carddet_sem);

	return 0;
}

/* CLOSE */
static int carddet_release(struct inode *inode, struct file *filp)
{
	struct carddet_dev *dev;

	DBG_MSG("** %s called  :  close", __func__);

	dev = filp->private_data;

	if (down_interruptible(&carddet_sem))
		return -ERESTARTSYS;

	dev->users--;
	DBG_MSG("   carddet: users=%d", dev->users);

	up(&carddet_sem);

	return 0;
}

/* IOCTL */
static int carddet_ioctl(struct file *filp, unsigned int cmd, unsigned long arg)
{
	struct carddet_dev *dev = (struct carddet_dev *)filp->private_data;
	int ret = 0;

	DBG_MSG("** %s called cmd(0x%08x) except(0x%08x)", __func__,
		cmd, CARD_DET_GET_INFO);

	if (_IOC_TYPE(cmd) != CARD_DET_IOC_MAGIC)
		return -ENOTTY;

	if (_IOC_NR(cmd) > CARD_DET_IOC_MAXNR)
		return -ENOTTY;

	switch (cmd) {
	case CARD_DET_GET_INFO:
		/* get card information */
		ret = carddet_ioc_get_info(dev, arg);
		break;

	case CARD_DET_CONTROL_POWER:
		/* control power */
		ret = carddet_ioc_control_power(dev, arg);
		break;

	case CARD_DET_GET_XDET:
		/* get card xdet status */
		ret = carddet_ioc_get_xdet(dev, arg);
		break;

	case CARD_DET_REQ_DETECT:
		/* request to detect a card */
		ret = carddet_ioc_req_detect(dev);
		break;

	default:
		ERR_MSG("** carddet_ioctl in OTHER:=0x%08x", cmd);
		ret = -ENOTTY;
		break;
	}

	return ret;
}

/* POLL */
static unsigned int carddet_poll(struct file *file, poll_table *wait)
{
	struct carddet_dev *dev = (struct carddet_dev *)file->private_data;
	int ret = 0;

	DBG_MSG("** %s called file_p=%08x wait_p=%08x", __func__,
		(unsigned int)file, (unsigned int)wait);

	poll_wait(file, &dev->wqh, wait);

#ifdef ENABLE_STAT_QUEUE
	if (!list_empty(&dev->list)) {
		{
			struct carddet_detect_status *status =
					list_first_entry(&dev->list,
					struct carddet_detect_status, entry);
			if (status)
				INFO_MSG("%s() status:%d,type:%d,ro:%d\n",
					 __func__, status->state,
					 status->type, status->ro);
		}

		ret |= POLLIN | POLLRDNORM;
	}
#else  /* ENABLE_STAT_QUEUE */
	if (dev->wake_up_flg != 0) {
		dev->wake_up_flg = 0;
		ret |= POLLIN | POLLRDNORM;
	}
#endif /* ENABLE_STAT_QUEUE */

	DBG_MSG("** carddet_poll: return(%d)", ret);

	return ret;
}

/**********************/
/*@ INTERRUPT HANDLER */
/**********************/
static irqreturn_t carddet_eint_handler(int irq, void *arg)
{
	struct msdc_host *host = carddet_devices->host_sd;
	int got_bad_card = 0;
	unsigned long flags;

	DBG_MSG("** %s called", __func__);

	spin_lock_irqsave(&carddet_devices->lock, flags);

	if (carddet_devices->suspend_flag == 0) {
		if (host->hw->cd_level ^ host->sd_cd_polarity) {
			got_bad_card = host->block_bad_card;
			host->card_inserted = 0;
			if (host->mmc && host->mmc->card) {
				DBG_MSG("** %s mmc_card_set_removed --called",
					 __func__);
				mmc_card_set_removed(host->mmc->card);
			}
		}
	}

	host->sd_cd_polarity = (~(host->sd_cd_polarity)) & 0x1;
	host->block_bad_card = 0;

	spin_unlock_irqrestore(&carddet_devices->lock, flags);

#ifdef CUST_EINT_MSDC1_INS_NUM
	/* control Eint irq */
	if (0 == host->sd_cd_polarity)
		mt_eint_set_polarity(CUST_EINT_MSDC1_INS_NUM, MT_POLARITY_LOW);
	else
		mt_eint_set_polarity(CUST_EINT_MSDC1_INS_NUM, MT_POLARITY_HIGH);
#endif

	if (carddet_devices->suspend_flag == 0 &&  got_bad_card == 0) {
		carddet_devices->thread_req = CARDDET_REQ_IRQ;
		wake_up_interruptible(&carddet_devices->thread_wq);
	}

	DBG_MSG("SD card %s(%x:%x)",
		(carddet_devices->host_sd->hw->cd_level ^
		 carddet_devices->host_sd->sd_cd_polarity) ?
			"insert" : "remove",
		 carddet_devices->host_sd->hw->cd_level,
		 carddet_devices->host_sd->sd_cd_polarity);

	return IRQ_HANDLED;
}

/****************/
/*@ MAIN THREAD */
/****************/
static int carddet_thread(void *arg)
{
	unsigned long flags;
	int ret;
	struct msdc_host *host_sd;

	for (;;) {
		if (carddet_devices->thread_timeout != 0) {
			/* waiting for events with timeout */
			ret = wait_event_interruptible_timeout(
					 carddet_devices->thread_wq,
					(carddet_devices->thread_req !=
					 CARDDET_REQ_NONE) ||
					 kthread_should_stop(),
					 carddet_devices->thread_timeout);
		} else {
			/* waiting for events */
			ret = wait_event_interruptible(
					 carddet_devices->thread_wq,
					(carddet_devices->thread_req !=
					 CARDDET_REQ_NONE) ||
					 kthread_should_stop());
		}
		if (ret == -ERESTARTSYS)
			continue;

		/* catch a stop event */
		if ((ret < 0) || kthread_should_stop())
			break;

		if (carddet_devices->suspend_flag) {
			carddet_devices->thread_req = CARDDET_REQ_NONE;
			carddet_devices->thread_seq = CARDDET_SEQ_NONE;
			carddet_devices->thread_timeout = CARDDET_TIMEOUT_IDLE;
			carddet_devices->insert_state_flag =
				CARDDET_STATE_FLAG_NONE;

			continue;
		}

		host_sd = carddet_devices->host_sd;
		DBG_MSG("Evnet:sequence:%d, request:%d, timeout:%d",
			carddet_devices->thread_seq,
			carddet_devices->thread_req, ret);

		if (carddet_devices->thread_req == CARDDET_REQ_IRQ) {
			/* Request: Detect/Remove a card
					(from Interrupt handler) */
			if (host_sd->hw->cd_level ^ host_sd->sd_cd_polarity) {
				carddet_devices->thread_seq =
					CARDDET_SEQ_PWRON;
				carddet_devices->thread_timeout =
					CARDDET_TIMEOUT_WAIT_PWR_STABLE;

				spin_lock_irqsave(&carddet_devices->lock,
							flags);
				carddet_devices->det_state =
					ICX_CARDDET_STATE_INSERTED;
				carddet_devices->read_only = 0;
				carddet_devices->insert_state_flag =
					CARDDET_STATE_FLAG_ALL;
				spin_unlock_irqrestore(&carddet_devices->lock,
							flags);

				carddet_control_power(
						carddet_devices->det_state);
				carddet_devices->thread_req = CARDDET_REQ_NONE;
			} else {
				carddet_card_removed();
				tasklet_hi_schedule(&host_sd->card_tasklet);
			}

			carddet_devices->retry_count = 0;
		} else {
			/* Timeout evnet */
			switch (carddet_devices->thread_seq) {
			case CARDDET_SEQ_PWRON:
				carddet_notify_status(carddet_devices);

				/* fall through */

			case CARDDET_SEQ_RETRY:
				/* translate sequence and timer */
				carddet_devices->thread_seq =
					CARDDET_SEQ_SET_PULL;
				carddet_devices->thread_timeout =
					CARDDET_TIMEOUT_WAIT_PULL_STABLE;

				/* get card type */
				spin_lock_irqsave(&carddet_devices->lock,
							flags);
				carddet_devices->card_type =
							ICX_CARDDET_TYPE_SD;
				spin_unlock_irqrestore(&carddet_devices->lock,
							flags);
				break;

			case CARDDET_SEQ_SET_PULL:
				if (enable_req_det_func) {
				    spin_lock_irqsave(
						&carddet_devices->lock, flags);
				    carddet_devices->insert_state_flag &=
						~CARDDET_STATE_FLAG_WAIT_STABLE;

				    if (carddet_devices->insert_state_flag) {
					spin_unlock_irqrestore(
						&carddet_devices->lock, flags);

					/* not ready to detect ->
							keep current state */
					carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_IDLE;
				    } else {
					spin_unlock_irqrestore(
						&carddet_devices->lock, flags);

					carddet_devices->thread_seq =
							CARDDET_SEQ_INIT;
					carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_WAIT_INIT;

					/* initialize SD/MS */
					if (carddet_devices->card_type ==
						ICX_CARDDET_TYPE_SD)
					    tasklet_hi_schedule(
						&host_sd->card_tasklet);
				    }
				} else {
					/* translate sequence and timer */
					carddet_devices->thread_seq =
						CARDDET_SEQ_INIT;
					carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_WAIT_INIT;

					/* initialize SD/MS */
					if (carddet_devices->card_type ==
							ICX_CARDDET_TYPE_SD)
					    tasklet_hi_schedule(
							&host_sd->card_tasklet);
				}
				break;

			case CARDDET_SEQ_INIT:
				/* get driver data */
				if (carddet_check_result()) {
					/* translate sequence and timer */
					carddet_devices->thread_seq =
						CARDDET_SEQ_RETRY;
					carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_RETRY;
				} else {
					/* translate sequence and
							timer to init value */
					carddet_devices->thread_seq =
						CARDDET_SEQ_NONE;
					carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_IDLE;
					carddet_devices->retry_count = 0;
				}
				break;

			default:
				/* translate sequence and timer to init value */
				carddet_devices->thread_seq = CARDDET_SEQ_NONE;
				carddet_devices->thread_timeout =
						CARDDET_TIMEOUT_IDLE;

				spin_lock_irqsave(&carddet_devices->lock,
							flags);
				carddet_devices->insert_state_flag =
						CARDDET_STATE_FLAG_NONE;
				spin_unlock_irqrestore(&carddet_devices->lock,
							flags);
				break;
			}

			if (carddet_devices->thread_req != CARDDET_REQ_IRQ)
				carddet_devices->thread_req = CARDDET_REQ_NONE;
		}
	}

	return 0;
}

/**************************/
/*@ DRIVER ENTRY ROUTINES */
/**************************/
/* SUSPEND */
static int carddet_suspend(struct platform_device *device, pm_message_t state)
{
	DBG_MSG("** %s called", __func__);

	carddet_devices->suspend_flag = 1;
	prev_det_state = carddet_devices->det_state;
	if (carddet_devices->det_state == ICX_CARDDET_STATE_DETECTED)
		carddet_card_removed();

	return 0;
}

/* RESUME */
static int carddet_resume(struct platform_device *device)
{
	DBG_MSG("** %s called", __func__);

	carddet_devices->suspend_flag = 0;

	if (carddet_get_hw_detect_state()) {
		DBG_MSG("** %s insert", __func__);
		carddet_devices->thread_req        = CARDDET_REQ_IRQ;
		carddet_devices->thread_seq        = CARDDET_SEQ_PWRON;
		carddet_devices->retry_count       = 0;
		carddet_devices->thread_timeout    = CARDDET_TIMEOUT_IDLE;
		wake_up_interruptible(&carddet_devices->thread_wq);
	} else if ((prev_det_state != ICX_CARDDET_STATE_REMOVED) &&
		   (!carddet_get_hw_detect_state())) {
		DBG_MSG("** %s remove (notify to user)", __func__);
		carddet_card_removed();
		tasklet_hi_schedule(&carddet_devices->host_sd->card_tasklet);
	} else if (carddet_devices->det_state != prev_det_state) {
		DBG_MSG("** %s remove (don't notify to user)", __func__);
		tasklet_hi_schedule(&carddet_devices->host_sd->card_tasklet);
	}

	return 0;
}

/* PROBE */
static int carddet_probe(void)
{
	dev_t dev = 0;
	int   ret;
	struct device *carddet_dev = NULL;

	DBG_MSG("** %s called", __func__);

	if (carddet_major != 0) {
		/* fixed */
		dev = MKDEV(carddet_major, carddet_minor);
		ret = register_chrdev_region(dev, 1, CARDDET_CDEV_FILE_NAME);
		DBG_MSG("### register_chrdev_region carddet: major=%d minor=%d",
			 carddet_major, carddet_minor);
	} else {
		/* dynamic */
		ret = alloc_chrdev_region(&dev, carddet_minor, 1,
					  CARDDET_CDEV_FILE_NAME);
		carddet_major = MAJOR(dev);
		DBG_MSG("### alloc_chrdev_region carddet: major=%d minor=%d\n",
			 carddet_major, carddet_minor);
	}

	if (ret < 0) {
		ERR_MSG("carddet: can't get major. (%d)", ret);
		return ret;
	}

	/*== device info ==*/
	carddet_devices = kmalloc(sizeof(struct carddet_dev), GFP_KERNEL);
	if (carddet_devices == NULL) {
		ret = -ENOMEM;
		goto err_kmalloc;
	}

	/*== initialize parameters ==*/
	spin_lock_init(&carddet_devices->lock);
	init_waitqueue_head(&carddet_devices->wqh);
	init_waitqueue_head(&carddet_devices->thread_wq);

#ifdef ENABLE_STAT_QUEUE
	INIT_LIST_HEAD(&carddet_devices->list);
#else  /* ENABLE_STAT_QUEUE */
	carddet_devices->wake_up_flg    = 0;
#endif /* ENABLE_STAT_QUEUE */

	carddet_devices->users          = 0;
	carddet_devices->retry_count    = 0;
	carddet_devices->card_type      = ICX_CARDDET_TYPE_SD;

	carddet_devices->thread_timeout = CARDDET_TIMEOUT_IDLE;
	carddet_devices->suspend_flag   = 0;

	carddet_devices->thread_req = CARDDET_REQ_NONE;
	carddet_devices->thread_seq = CARDDET_SEQ_NONE;
	carddet_devices->insert_state_flag = CARDDET_STATE_FLAG_NONE;
	carddet_devices->det_state  = ICX_CARDDET_STATE_REMOVED;
	carddet_control_power(carddet_devices->det_state);

	/*== initialize cdev ==*/
	cdev_init(&carddet_devices->cdev, &carddet_fops);
	/* carddet_devices->cdev = cdev_alloc(); */
	carddet_devices->cdev.owner = THIS_MODULE;
	carddet_devices->cdev.ops   = &carddet_fops;

	ret = cdev_add(&carddet_devices->cdev, dev, 1);
	if (ret) {
		ERR_MSG("carddet: error(%d) adding carddet", ret);
		goto err_cdev_add;
	}

	carddet_class = class_create(THIS_MODULE, "icx_carddet");

	if (IS_ERR(carddet_class)) {
		ERR_MSG("carddet: error(%d) class_create", ret);
		goto err_cdev_add;
	}

	carddet_dev = device_create(carddet_class, NULL,
				    dev, NULL, "icx_carddet");
	if (IS_ERR(carddet_dev))
		goto err_cdev_add;

	/*== create task ==*/
	carddet_devices->thread_task =
		kthread_create(carddet_thread, carddet_devices, MODULE_NAME);
	if (IS_ERR(carddet_devices->thread_task)) {
		ret = PTR_ERR(carddet_devices->thread_task);
		ERR_MSG("carddet: error(%d) failed to create a thread", ret);
		goto err_create_thread;
	}
	wake_up_process(carddet_devices->thread_task);

	/*== carddet function register ==*/
	msdc_set_carddet_func(&carddet_func);
	msc_bind_carddet(&carddet_to_msc_func_tbl);

	/*== request Eint IRQ ==*/
	carddet_devices->host_sd = mtk_msdc_host[1];
	msdc_eint_register(&carddet_ops, true);

	DBG_MSG("carddet: probe end");

	return 0;

	/*== Error Case  ==*/
err_create_thread:
	/* do nothing */

err_cdev_add:
	kfree(carddet_devices);

err_kmalloc:
	unregister_chrdev_region(dev, 1);

	return ret;
}

/* REMOVE */
static void carddet_remove(void)
{
	dev_t dev;

	DBG_MSG("carddet: remove");

#ifdef CUST_EINT_MSDC1_INS_NUM
	/* disable eint */
	msdc_eint_register(&carddet_ops, false);
#endif /* CUST_EINT_MSDC1_INS_NUM */

	/* get device file */
	dev = MKDEV(carddet_major, carddet_minor);

	kthread_stop(carddet_devices->thread_task);

	if (carddet_devices) {
#ifdef ENABLE_STAT_QUEUE
		struct carddet_detect_status *status, *status_next;
		list_for_each_entry_safe(status, status_next,
					 &carddet_devices->list, entry) {
			list_del(&status->entry);
			kfree(status);
		}
#endif /* ENABLE_STAT_QUEUE */

		device_destroy(carddet_class, dev);
		class_destroy(carddet_class);
		cdev_del(&carddet_devices->cdev);
		kfree(carddet_devices);
	}

	unregister_chrdev_region(dev, 1);

	DBG_MSG("carddet: remove end");
}

/* SHUTDOWN */
static void carddet_shutdown(struct platform_device *device)
{
	DBG_MSG("** %s called", __func__);

	carddet_remove();
}

static struct platform_driver carddet_driver = {
	.remove   = carddet_remove,
	.probe    = carddet_probe,
	.suspend  = carddet_suspend,
	.resume   = carddet_resume,
	.shutdown = carddet_shutdown,
	.driver   = {
		.name = MODULE_NAME,
	},
};

/* INIT */
static int __init carddet_init(void)
{
	INFO_MSG("Starting ICX card detect driver.");

	prev_det_state = 0;
	enable_req_det_func = 0;
	carddet_class = NULL;

	return platform_driver_register(&carddet_driver);
}

/* EXIT */
static void __exit carddet_exit(void)
{
	INFO_MSG("Exiting ICX card detect driver.");

	platform_driver_unregister(&carddet_driver);
}

module_init(carddet_init);
module_exit(carddet_exit);

MODULE_AUTHOR("sony corporation");
MODULE_DESCRIPTION("ICX card detect driver");
MODULE_LICENSE("GPL");
/* MODULE_LICENSE("proprietary"); */

